/*
	Minimalistic FAT32 library
	Copyright (C) 2009  Avelino Herrera Morales

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "physical.h"
#include "fat32.h"


uint32_t fatGetClusterLBA(fatUnit *fat, uint32_t cluster) {
	return fat->clusterBeginLBA + (cluster - 2) * fat->sectorsPerCluster;
}


uint32_t fatGetNextCluster(fatUnit *fat, uint32_t cluster) {
	uint32_t lba, lbaOffset, ret;
	if ((cluster < 2) || (cluster >= 0x0FFFFFF8))
		return 0x0FFFFFFF;
	lba = fat->fatBeginLBA + (cluster / 128);
	lbaOffset = (cluster % 128) * 4;
	deviceReadSector(lba);
	memcpy(&ret, currentSector + lbaOffset, 4);
	return ret;
}


void fatSetClusterChain(fatUnit *fat, uint32_t cluster, uint32_t nextCluster) {
	uint32_t lba, lbaOffset;
	if ((cluster < 2) || (cluster >= 0x0FFFFFF8))
		return;
	// write FAT1
	lba = fat->fatBeginLBA + (cluster / 128);
	lbaOffset = (cluster % 128) * 4;
	deviceReadSector(lba);
	memcpy(currentSector + lbaOffset, &nextCluster, 4);
	deviceWriteSector(lba);
	// write FAT2
	lba = fat->fat2BeginLBA + (cluster / 128);
	deviceReadSector(lba);
	memcpy(currentSector + lbaOffset, &nextCluster, 4);
	deviceWriteSector(lba);
	printf("fatSetClusterChain: 0x%08X ---> 0x%08X\n", cluster, nextCluster);
}


uint32_t fatGetFreeCluster(fatUnit *fat) {
	uint32_t i, clusterRel, clusterAbs, aux;
	clusterRel = clusterAbs = 2;
	for (i = 0; i < fat->sectorsPerFat; i++) {
		deviceReadSector(fat->fatBeginLBA + i);
		for ( ; clusterRel < 128; clusterRel++, clusterAbs++) {
			memcpy(&aux, currentSector + (clusterRel * 4), 4);
			if (aux == 0)
				return clusterAbs;
		}
		clusterRel = 0;
	}
	return 0;
}


void doTab(uint16_t tab) {
	uint16_t i;
	for (i = 0; i < tab; i++)
		printf("\t");
}


uint32_t fatShowDirectoryEntry(uint16_t tab, uint8_t *p) {
	uint32_t cluster;
	uint16_t cluster_hi, cluster_lo;
	uint32_t size;
	uint16_t i;
	doTab(tab);
	printf("'");
	for (i = 0; i < 11; i++)
		printf("%c", p[i]);
	printf("' ");
	if ((p[0x0B] & 0x01) != 0)
		printf("readonly ");
	if ((p[0x0B] & 0x02) != 0)
		printf("hidden ");
	if ((p[0x0B] & 0x04) != 0)
		printf("system ");
	if ((p[0x0B] & 0x08) != 0)
		printf("volumeid ");
	if ((p[0x0B] & 0x10) != 0)
		printf("directory ");
	if ((p[0x0B] & 0x20) != 0)
		printf("archive ");
	memcpy(&cluster_hi, &p[0x14], 2);
	memcpy(&cluster_lo, &p[0x1A], 2);
	cluster = (((uint32_t) cluster_hi) << 16) | ((uint32_t) cluster_lo);
	printf(" cluster=0x%08X", cluster);
	memcpy(&size, &p[0x1C], 4);
	printf(" size=%u", size);
	return cluster;
}


void fatShowDirectoryEntries(fatUnit *fat, uint16_t tab, uint32_t cluster) {
	uint8_t *p, i, clusterSector;
	uint32_t newCluster;
	uint8_t privateCurrentSector[SECTOR_SIZE];
	uint32_t sector;
	uint32_t nextCluster;
	while (1) {
		doTab(tab);
		printf("[0x%08X]\n", cluster);
		sector = fatGetClusterLBA(fat, cluster);
		for (clusterSector = 0; clusterSector < fat->sectorsPerCluster; clusterSector++, sector++) {
			deviceReadSector(sector);
			memcpy(privateCurrentSector, currentSector, SECTOR_SIZE);
			for (i = 0, p = privateCurrentSector; i < 16; i++, p += 32) {
				if (*p == 0x00) {
					// end of directory
					return;
				}
				if (*p == 0xE5) {
					// unused entry (deleted)
					continue;
				}
				if ((p[0x0B] & 0x0F) == 0x0F) {
					// long filename entry
					continue;
				}
				newCluster = fatShowDirectoryEntry(tab, p);
				printf("\n");
				if (((p[0x0B] & 0x10) != 0) && (*p != '.')) {
					// if directory call recursive
					fatShowDirectoryEntries(fat, tab + 1, newCluster);
				}
			}
		}
		nextCluster = fatGetNextCluster(fat, cluster);
		//printf("nextCluster = 0x%08X\n", nextCluster);
		if (nextCluster >= 0x0FFFFFF8)
			break;
		else
			cluster = nextCluster & 0x0FFFFFFF;
	}
}


char *getCharPtr(char *s, char c) {
	while (*s != 0) {
		if (*s == c)
			break;
		s++;
	}
	return s;
}


char toUpper(char c) {
	if ((c < 'a') || (c > 'z'))
		return c;
	return 'A' + (c - 'a');
}


uint32_t fatSearchChildFileCluster(fatUnit *fat, uint32_t parentCluster, char *name, uint8_t *isDirectory, uint32_t *size, uint32_t *entrySector, uint8_t *entryOffset) {
	uint32_t nextParentCluster, sector, retCluster;
	uint8_t clusterSector;
	uint16_t clusterHi, clusterLo;
	uint8_t i, *p;
	while (1) {
		sector = fatGetClusterLBA(fat, parentCluster);
		for (clusterSector = 0; clusterSector < fat->sectorsPerCluster; clusterSector++, sector++) {
			deviceReadSector(sector);
			for (i = 0, p = currentSector; i < 16; i++, p += 32) {
				if (*p == 0x00) {
					// end of directory
					return 0;
				}
				if (*p == 0xE5) {
					// unused entry (deleted)
					continue;
				}
				if ((p[0x0B] & 0x0F) == 0x0F) {
					// long filename entry
					continue;
				}
				if (memcmp(name, p, 11) == 0) {
					if (((p[0x0B] & 0x10) != 0) && (*p != '.'))
						*isDirectory = 1;
					else
						*isDirectory = 0;
					memcpy(size, &p[0x1C], 4);
					memcpy(&clusterHi, &p[0x14], 2);
					memcpy(&clusterLo, &p[0x1A], 2);
					*entrySector = sector;
					*entryOffset = ((uint32_t) i) * 32;
					retCluster = (((uint32_t) clusterHi) << 16) | ((uint32_t) clusterLo);
					return retCluster;
				}
			}
		}
		nextParentCluster = fatGetNextCluster(fat, parentCluster);
		if (nextParentCluster >= 0x0FFFFFF8)
			break;
		else
			parentCluster = nextParentCluster & 0x0FFFFFFF;
	}
	return 0;
}


uint32_t fatGetLastCluster(fatUnit *fat, fatFile *file) {
	uint32_t lba, lbaOffset, newCluster, cluster;
	cluster = file->startCluster;
	if ((cluster < 2) || (cluster >= 0x0FFFFFF8))
		return 0x0FFFFFFF;
	while (1) {
		lba = fat->fatBeginLBA + (cluster / 128);
		lbaOffset = (cluster % 128) * 4;
		deviceReadSector(lba);
		memcpy(&newCluster, currentSector + lbaOffset, 4);
		if ((newCluster < 2) || (newCluster >= 0x0FFFFFF8))
			return cluster;
		else
			cluster = newCluster;
	}
	return 0x0FFFFFFF;
}


void fatFileSetStartCluster(fatUnit *fat, fatFile *file, uint32_t cluster) {
	uint16_t clusterHi = (uint16_t) (cluster >> 16);
	uint16_t clusterLo = (uint16_t) (cluster & 0x0000FFFF);
	deviceReadSector(file->entrySector);
	memcpy(currentSector + file->entryOffset + 0x14, &clusterHi, 2);
	memcpy(currentSector + file->entryOffset + 0x1A, &clusterLo, 2);
	deviceWriteSector(file->entrySector);
	file->startCluster = cluster;
}


void fatFileSetSize(fatUnit *fat, fatFile *file, uint32_t newSize) {
	char aux[12];
	printf("fatFileSetSize: %u\n", newSize);
	printf("fatFileSetSize: file entry at sector %u offset %u\n", file->entrySector, file->entryOffset);
	deviceReadSector(file->entrySector);
	memcpy(aux, currentSector + file->entryOffset, 11);
	aux[11] = 0;
	printf("fatFileSetSize: this should be the file name ==> '%11s'\n", aux);
	memcpy(currentSector + file->entryOffset + 0x1C, &newSize, 4);
	deviceWriteSector(file->entrySector);
	file->size = newSize;
}


uint8_t fatAdjustSize(fatUnit *fat, fatFile *file, uint32_t newSize) {
	uint32_t bytesPerCluster = SECTOR_SIZE * ((uint32_t) fat->sectorsPerCluster);
	uint32_t clustersCurrentSize = file->size / bytesPerCluster;
	if ((file->size % bytesPerCluster) > 0)
		clustersCurrentSize++;
	uint32_t clustersNewSize = newSize / bytesPerCluster;
	if ((newSize % bytesPerCluster) > 0)
		clustersNewSize++;
	printf("fatAdjustSize: clustersCurrentSize=%u clustersNewSize=%u\n", clustersCurrentSize, clustersNewSize);
	if (clustersNewSize > clustersCurrentSize) {
		// need more clusters
		uint32_t clustersNeeded = clustersNewSize - clustersCurrentSize;
		uint32_t freeCluster;
		uint32_t lastCluster = fatGetLastCluster(fat, file);
		printf("fatAdjustSize: adding %u clusters\n", clustersNeeded);
		while (clustersNeeded--) {
			freeCluster = fatGetFreeCluster(fat);
			if (freeCluster == 0) {
				if (lastCluster < 0x0FFFFFF8)
					fatSetClusterChain(fat, lastCluster, 0x0FFFFFFF);
				printf("fatAdjustSize: disk full\n");
				return 0;
			}
			if (lastCluster >= 0x0FFFFFF8)
				fatFileSetStartCluster(fat, file, freeCluster);
			else
				fatSetClusterChain(fat, lastCluster, freeCluster);
			lastCluster = freeCluster;
		}
		fatSetClusterChain(fat, lastCluster, 0x0FFFFFFF);
		printf("fatAdjustSize: set new size to %u\n", newSize);
	}
	else if (clustersNewSize < clustersCurrentSize) {
		// need less clusters
		// TODO (is this needed?)
		uint32_t clustersFreed = clustersCurrentSize - clustersNewSize;
		printf("fatAdjustSize: removing %u clusters (not implemented)\n", clustersFreed);
		return 0;
	}
	fatFileSetSize(fat, file, newSize);
	return 1;
}


uint32_t fatWrite(fatUnit *fat, fatFile *file, uint32_t offset, uint32_t size, uint8_t *source) {
	uint32_t bytesPerCluster = SECTOR_SIZE * ((uint32_t) fat->sectorsPerCluster);
	uint32_t clusterIndex;
	uint32_t clusterOffset;
	uint32_t cluster;
	uint32_t byteCount = 0;
	uint32_t sectorIndex;
	uint32_t sectorOffset;
	uint32_t sector;
	if (size == 0)
		return 0;
	if ((offset > file->size) || (file->status != FAT_FILE_STATUS_OK))
		return 0;
	if ((offset + size) > file->size) {
		printf("fatWrite: need to adjust size from %u to %u\n", file->size, offset + size);
		if (fatAdjustSize(fat, file, offset + size) == 0)
			return 0;
	}
	clusterIndex = offset / bytesPerCluster;
	clusterOffset = offset % bytesPerCluster;
	cluster = file->startCluster;
	while (clusterIndex--) {
		cluster = fatGetNextCluster(fat, cluster);
		//printf(" %08X", cluster);
	}
	if (cluster >= 0x0FFFFFF8)
		return 0;
	sectorIndex = clusterOffset / SECTOR_SIZE;
	sectorOffset = clusterOffset % SECTOR_SIZE;
	printf("fatWrite: sectorIndex=%u sectorOffset=%u\n", sectorIndex, sectorOffset);
	while (byteCount < size) {
		if (cluster >= 0x0FFFFFF8)
			break;
		sector = fatGetClusterLBA(fat, cluster);
		printf("fatWrite: cluster 0x%08X starts at sector %u\n", cluster, sector);
		while ((byteCount < size) && (sectorIndex < fat->sectorsPerCluster)) {
			uint16_t i;
			printf("fatWrite: reading sector %u from disk\n", sector + sectorIndex);
			deviceReadSector(sector + sectorIndex);
			for (i = sectorOffset; (i < SECTOR_SIZE) && (byteCount < size); i++, byteCount++)
				currentSector[i] = *source++;
			printf("fatWrite: writing modified sector %u to disk\n", sector + sectorIndex);
			deviceWriteSector(sector + sectorIndex);
			sectorOffset = 0;
			sectorIndex++;
		}
		sectorIndex = 0;
		cluster = fatGetNextCluster(fat, cluster);
	}
	return byteCount;
}


uint32_t fatRead(fatUnit *fat, fatFile *file, uint32_t offset, uint32_t size, uint8_t *dest) {
	uint32_t sector;
	uint32_t cluster;
	uint32_t clusterIndex;
	uint32_t clusterOffset;
	uint32_t sectorIndex;
	uint32_t sectorOffset;
	uint32_t byteCount = 0;
	uint32_t bytesPerCluster = SECTOR_SIZE * ((uint32_t) fat->sectorsPerCluster);
	if ((offset + size) > file->size) {
		if (offset >= file->size)
			size = 0;
		else
			size = file->size - offset;
	}
	//printf("fatRead: bytesPerCluster=%u\n", bytesPerCluster);
	if ((file->status != FAT_FILE_STATUS_OK) || (size == 0))
		return 0;
	clusterIndex = offset / bytesPerCluster;
	clusterOffset = offset % bytesPerCluster;
	//printf("fatRead: clusterIndex=%u clusterOffset=%u\n", clusterIndex, clusterOffset);
	cluster = file->startCluster;
	//printf("fatRead: cluster chain: 0x%08X", cluster);
	while (clusterIndex--) {
		cluster = fatGetNextCluster(fat, cluster);
		//printf(" %08X", cluster);
	}
	//printf("\n");
	if (cluster >= 0x0FFFFFF8)
		return 0;
	sectorIndex = clusterOffset / SECTOR_SIZE;
	sectorOffset = clusterOffset % SECTOR_SIZE;
	//printf("fatRead: sectorIndex=%u sectorOffset=%u\n", sectorIndex, sectorOffset);
	while (byteCount < size) {
		if (cluster >= 0x0FFFFFF8)
			break;
		sector = fatGetClusterLBA(fat, cluster);
		//printf("fatRead: cluster 0x%08X starts at sector %u\n", cluster, sector);
		while ((byteCount < size) && (sectorIndex < fat->sectorsPerCluster)) {
			uint16_t i;
			//printf("fatRead: reading sector %u from disk\n", sector + sectorIndex);
			deviceReadSector(sector + sectorIndex);
			for (i = sectorOffset; (i < SECTOR_SIZE) && (byteCount < size); i++, byteCount++)
				*dest++ = currentSector[i];
			sectorOffset = 0;
			sectorIndex++;
		}
		sectorIndex = 0;
		cluster = fatGetNextCluster(fat, cluster);
	}
	return byteCount;
}


void fatOpen(fatUnit *fat, fatFile *file, char *path) {
	uint32_t parentCluster = fat->rootDirectoryCluster;
	uint32_t childCluster;
	char *currentPathToken = path;
	uint8_t isDirectory;
	uint32_t size;
	uint32_t entrySector;
	uint8_t entryOffset;
	strcpy(file->path, path);
	while (1) {
		char *nextPathToken = getCharPtr(currentPathToken, '/');
		//printf("fatOpen: currentPathToken='%s'  nextPathToken='%s'\n", currentPathToken, nextPathToken);
		childCluster = fatSearchChildFileCluster(fat, parentCluster, currentPathToken, &isDirectory, &size, &entrySector, &entryOffset);
		//printf("fatOpen: parentCluster=0x%08X childCluster=0x%08X\n", parentCluster, childCluster);
		if (childCluster == 0) {
			file->status = FAT_FILE_STATUS_NOT_FOUND;
			return;
		}
		if (isDirectory == 1) {
			//printf("fatOpen: isDirectory == 1\n");
			if (*nextPathToken == '/') {
				currentPathToken = nextPathToken + 1;
				parentCluster = childCluster;
				continue;
			}
			else {
				file->status = FAT_FILE_STATUS_OK_DIR;
				file->size = 0;
				file->startCluster = childCluster;
				file->entrySector = entrySector;
				file->entryOffset = entryOffset;
				return;
			}
		}
		else {
			//printf("fatOpen: isDirectory == 0\n");
			if (*nextPathToken == '/') {
				file->status = FAT_FILE_STATUS_NOT_FOUND;
				return;
			}
			else {
				file->status = FAT_FILE_STATUS_OK;
				file->size = size;
				file->startCluster = childCluster;
				file->entrySector = entrySector;
				file->entryOffset = entryOffset;
				printf("fatOpen: file entry at sector %u offset %u\n", file->entrySector, file->entryOffset);
				return;
			}
		}
	}
}


uint8_t fatDelete(fatUnit *fat, char *path) {
	fatFile file;
	fatOpen(fat, &file, path);
	if (file.status == FAT_FILE_STATUS_OK) {
		deviceReadSector(file.entrySector);
		*(currentSector + file.entryOffset) = 0xE5;
		deviceWriteSector(file.entrySector);
		return 1;
	}
	else if (file.status == FAT_FILE_STATUS_OK_DIR) {
		// TODO delete directory
		return 0;
	}
	return 0;
}


void fatCreate(fatUnit *fat, fatFile *file, char *folderPath, char *fileName) {
	uint32_t cluster, nextCluster;
	uint32_t sector;
	uint8_t i, endOfDirectoryFound = 0, deletedFound = 0;
	uint8_t *p;
	uint8_t clusterSector;
	if (*fileName == 0) {
		printf("fatCreate: empty file name\n");
		file->status = FAT_FILE_STATUS_NOT_FOUND;
		return;
	}
	// search the folder cluster
	if (*folderPath == 0) {
		printf("fatCreate: empty folder path ---> root directory\n");
		cluster = fat->rootDirectoryCluster;
	}
	else {
		fatOpen(fat, file, folderPath);
		if (file->status != FAT_FILE_STATUS_OK_DIR) {
			file->status = FAT_FILE_STATUS_NOT_FOUND;
			return;
		}
		cluster = file->startCluster;
	}
	// find end-of-directory entry
	while (1) {
		printf("fatCreate: searching in cluster 0x%08X\n", cluster);
		sector = fatGetClusterLBA(fat, cluster);
		for (clusterSector = 0; clusterSector < fat->sectorsPerCluster; clusterSector++, sector++) {
			deviceReadSector(sector);
			for (i = 0, p = currentSector; i < 16; i++, p += 32) {
				if (*p == 0x00) {
					// end of directory
					printf("fatCreate: end-of-directory entry found\n");
					endOfDirectoryFound = 1;
					break;
				}
				if (*p == 0xE5) {
					// deleted entry
					printf("fatCreate: deleted entry found\n");
					deletedFound = 1;
					break;
				}
			}
			if (endOfDirectoryFound || deletedFound)
				break;
		}
		if (endOfDirectoryFound || deletedFound)
			break;
		nextCluster = fatGetNextCluster(fat, cluster);
		//printf("next_cluster = 0x%08X\n", next_cluster);
		if (nextCluster >= 0x0FFFFFF8)
			break;
		else
			cluster = nextCluster & 0x0FFFFFFF;
	}
	if (!endOfDirectoryFound && !deletedFound) {
		printf("fatCreate: neither end-of-directory nor deleted entry found\n");
		file->status = FAT_FILE_STATUS_NOT_FOUND;
		return;
	}
	// fill the file entry at end-of-directory entry
	printf("fatCreate: overwriting entry\n");
	memcpy(p, fileName, 11);   // file name
	*(p + 0x0B) = 0x20;        // attribute = archive
	*(p + 0x15) = 0x0F;        // start cluster = 0x0FFFFFFF
	*(p + 0x14) = 0xFF;
	*(p + 0x1B) = 0xFF;
	*(p + 0x1A) = 0xFF;
	memset(p + 0x1C, 0, 4);    // size = 0 bytes
	file->status = FAT_FILE_STATUS_OK;
	file->size = 0;
	file->startCluster = 0x0FFFFFFF;
	file->entrySector = sector;
	file->entryOffset = i;
	strcpy(file->path, folderPath);
	strcat(file->path, "/");
	strcat(file->path, fileName);
	if (deletedFound) {
		printf("fatCreate: deleted entry ---> write sector and done\n");
		deviceWriteSector(sector);
		return;
	}
	if (i < 15) {
		// still into the same sector
		printf("fatCreate: end-of-directory entry ---> still into the same sector\n");
		p += 32;
		*p = 0x00;
		deviceWriteSector(sector);
	}
	else {
		// must go to the next sector
		printf("fatCreate: end-of-directory entry ---> must go to the next sector\n");
		deviceWriteSector(sector);
		if (clusterSector < (fat->sectorsPerCluster - 1)) {
			// still into the same cluster
			printf("fatCreate: end-of-directory entry ---> still into the same cluster\n");
			deviceReadSector(sector + 1);
			*currentSector = 0x00;
			deviceWriteSector(sector + 1);
		}
		else {
			// must go to the next cluster
			printf("fatCreate: end-of-directory entry ---> must alloc a free cluster to put the new end-of-directory entry\n");
			nextCluster = fatGetFreeCluster(fat);
			if (nextCluster == 0) {
				printf("fatCreate: end-of-directory entry ---> disk full!\n");
				file->status = FAT_FILE_STATUS_NOT_FOUND;
				*p = 0x00;      // restore end-of-directory entry
				return;
			}
			else {
				printf("fatCreate: end-of-directory entry ---> allocating new cluster and chaining  0x%08X --> 0x%08X --> 0x0FFFFFFF\n", cluster, nextCluster);
				fatSetClusterChain(fat, cluster, nextCluster);
				fatSetClusterChain(fat, nextCluster, 0x0FFFFFFF);
				sector = fatGetClusterLBA(fat, cluster);
				deviceReadSector(sector);
				*currentSector = 0x00;
				deviceWriteSector(sector);
			}
		}
	}
}


uint8_t fatReadPartition(fatUnit *fat) {
	uint16_t bytesPerSector;
	uint8_t numFats;
	uint16_t signature;
	uint16_t reservedSectors;
	deviceReadSector(fat->firstLBA);
	memcpy(&bytesPerSector, &currentSector[0x0B], 2);
	printf("fatReadPartition: bytesPerSector=%d\n", bytesPerSector);
	if (bytesPerSector != SECTOR_SIZE) {
		printf("must be %u!\n", SECTOR_SIZE);
		return 0;
	}
	fat->sectorsPerCluster = currentSector[0x0D];
	printf("fatReadPartition: sectorsPerCluster=%d\n", fat->sectorsPerCluster);
	memcpy(&reservedSectors, &currentSector[0x0E], 2);
	printf("fatReadPartition: reservedSectors=%d\n", reservedSectors);
	numFats = currentSector[0x10];
	printf("fatReadPartition: numFats=%d\n", numFats);
	if (numFats != 2) {
		printf("must be 2!\n");
		return 0;
	}
	memcpy(&fat->sectorsPerFat, &currentSector[0x24], 4);
	printf("fatReadPartition: sectorsPerFat=%d\n", fat->sectorsPerFat);
	memcpy(&fat->rootDirectoryCluster, &currentSector[0x2C], 4);
	printf("fatReadPartition: rootDirectoryCluster=0x%08X\n", fat->rootDirectoryCluster);
	if (fat->rootDirectoryCluster != 2)
		printf("should be 2 (it is not an error)\n");
	memcpy(&signature, &currentSector[0x1FE], 2);
	printf("fatReadPartition: signature=0x%04X\n", signature);
	if (signature != 0xAA55) {
		printf("must be 0xAA55!\n");
		return 0;
	}
	fat->fatBeginLBA = fat->firstLBA + ((uint32_t) reservedSectors);
	fat->fat2BeginLBA = fat->fatBeginLBA + fat->sectorsPerFat;
	fat->clusterBeginLBA = fat->firstLBA + ((uint32_t) reservedSectors) + (2 * fat->sectorsPerFat);
	return 1;
}


uint8_t fatMount(fatUnit *fat) {
	uint8_t *p;
	uint8_t part;
	deviceReadSector(0);
	p = currentSector + 446;
	for (part = 0; part < 4; part++, p += 16) {
		printf("partition %d\n", part);
		if ((p[0] != 0x80) && (p[0] != 0x00)) {
			printf("\tnon valid\n");
			continue;
		}
		printf("\tactive");
		if (p[0] == 0x80)
			printf(" bootable");
		printf("\n");
		memcpy(&fat->firstLBA, &p[8], 4);
		memcpy(&fat->lastLBA, &p[12], 4);
		printf("\tfirst sector at %08x\n", fat->firstLBA);
		printf("\tlast  sector at %08x\n", fat->lastLBA);
		return fatReadPartition(fat);
		/*
		fat_read_root_directory();
		return 1;
		*/
	}
	return 0;
}

